/*
 * Copyright (c) 2019 gematik - Gesellschaft für Telematikanwendungen der Gesundheitskarte mbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.gematik.ti.vauchannel.protocol.helpers;

import org.bouncycastle.asn1.x509.SubjectPublicKeyInfo;
import org.bouncycastle.cert.X509CertificateHolder;
import org.bouncycastle.cert.jcajce.JcaX509CertificateHolder;
import org.bouncycastle.cert.ocsp.*;
import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.bouncycastle.operator.DigestCalculatorProvider;
import org.bouncycastle.operator.bc.BcDigestCalculatorProvider;
import org.bouncycastle.operator.jcajce.JcaContentSignerBuilder;

import java.security.PrivateKey;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.time.ZonedDateTime;
import java.util.Date;

public class OCSPResponseGenerator {
    public static OCSPResp gen(X509Certificate eeCertJca,
                               X509Certificate caCertJca,
                               PrivateKey caPrivateKey,
                               CertificateStatus status) throws CertificateException {
        return gen(eeCertJca, caCertJca, caPrivateKey, status, ZonedDateTime.now());
    }

    public static OCSPResp gen(X509Certificate eeCertJca,
                               X509Certificate caCertJca,
                               PrivateKey caPrivateKey,
                               CertificateStatus status,
                               ZonedDateTime dateTime) throws CertificateException {
        try {
            X509CertificateHolder caCert = new JcaX509CertificateHolder(caCertJca);

            DigestCalculatorProvider digCalcProv = new BcDigestCalculatorProvider();
            BasicOCSPRespBuilder basicBuilder = new BasicOCSPRespBuilder(
                    SubjectPublicKeyInfo.getInstance(caCertJca.getPublicKey().getEncoded()),
                    digCalcProv.get(CertificateID.HASH_SHA1));

            CertificateID certId = new CertificateID(digCalcProv.get(CertificateID.HASH_SHA1),
                    caCert, eeCertJca.getSerialNumber());

            basicBuilder.addResponse(certId, status);

            X509CertificateHolder[] chain = {new X509CertificateHolder(caCertJca.getEncoded())};
            BasicOCSPResp resp = basicBuilder.build(
                    new JcaContentSignerBuilder("SHA256withECDSA").setProvider(BouncyCastleProvider.PROVIDER_NAME).build(caPrivateKey),
                    chain, new Date(dateTime.toInstant().toEpochMilli()));

            OCSPRespBuilder builder = new OCSPRespBuilder();
            return builder.build(OCSPRespBuilder.SUCCESSFUL, resp);
        } catch (Exception e) {
            throw new CertificateException("cannot generate OCSP response", e);
        }
    }

}
